datatype Expr = Const(int)
              | Var(string)
              | Node(op: Op, args : List<Expr>)

datatype Op = Add | Mul | Sub | Div

datatype List<T> = Nil | Cons(head: T, tail: List<T>)

function Eval(e: Expr,  env: map<string, int>): int 
requires goodEnv(e, env)
{
    match e
    case Const(c) => c
    case Var(s) => env[s]
    case Node(op, args) => EvalList(args, op, env)
}

function EvalList(args: List<Expr>, op: Op, env: map<string, int>) : int
requires goodEnv'(args, env)
{
    match args 
    case Nil => 
      (match op case Add => 0 case Sub => 0 case Mul => 1 case Div => 1)
    case Cons(e, tail) =>
      var v0, v1 := Eval(e, env), EvalList(tail, op, env);
      match op 
      case Add => v0 + v1
      case Sub => v0 - v1
      case Mul => v0 * v1
      case Div => if v1 != 0 then v0 / v1 else 0
}


predicate goodEnv(e: Expr, env: map<string, int>) {
    match e
      case Const(_) => true
      case Var(name) => name in env.Keys
      case Node(op, args) => goodEnv'(args, env)
}

predicate goodEnv'(exprs: List<Expr>, env: map<string, int>) {
    match exprs
      case Nil => true
      case Cons(head, tail) => goodEnv(head, env) && goodEnv'(tail, env)
}





lemma {:axiom} envsGood(e: Expr, env1: map<string, int>, env2: map<string, int>)
requires env1.Keys == env2.Keys && goodEnv(e, env1)
ensures goodEnv(e, env2)
/*
{
     match e
      case Const(_) => assert(goodEnv(e, env2));
      case Var(name) => assert((goodEnv(e, env1) && name in env1.Keys) ==> name in env2.Keys ==> goodEnv(e, env2));
      case Node(op, Cons(head, tail)) => envsGood(head, env1, env2);
      case Node(_, Nil) => assert(true);
}
 */


method Main() {
    // we create the following expression: 10 * (y + (5 - x))
    var e1 := Node(Mul, 
                   Cons(Const(10), 
                        Cons(Node(Add, 
                              Cons(Var("y"),
                                    Cons(Node(Sub,
                                         Cons(Const(5),
                                         Cons(Var("x"),
                                         Nil))), Nil))), Nil)));

    var env : map<string, int> := map["x":= 3, "y":= 10];

    var res := Eval(e1, env);
    print(res);
    print("\n");
    assert(goodEnv(e1, env));

    var env2 := env["x":= 2];
    envsGood(e1, env, env2);
    assert(goodEnv(e1, env2));
    res := Eval(e1, env2);
    print(res);
    print("\n");
}
